@@ -3,14 +3,13 @@
 #
 # CVS Information:
 #     $Source: /home/globdev/CVS/globus-packages/gram/jobmanager/scripts/JobManager.pm,v $
-#     $Date: 2007/12/01 00:48:54 $
-#     $Revision: 1.4 $
-#     $Author: feller $
+#     $Date: 2010/07/07 20:59:36 $
+#     $Revision: 1.7.2.9 $
+#     $Author: bester $
 #
 use Globus::GRAM::Error;
 use Globus::GRAM::JobState;
 use Globus::GRAM::JobSignal;
-use Globus::GRAM::ExtensionsHandler;
 use Globus::Core::Paths;
 
 use POSIX;
@@ -20,7 +19,7 @@ use File::Copy;
 
 package Globus::GRAM::JobManager;
 
-my $cache_pgm = "$Globus::Core::Paths::bindir/globus-gass-cache";
+my $cache_pgm = "$Globus::Core::Paths::libexecdir/globus-gass-cache-util";
 my $url_copy_pgm = "$Globus::Core::Paths::bindir/globus-url-copy";
 my $info_pgm = "$Globus::Core::Paths::bindir/grid-proxy-info";
 
@@ -86,32 +85,31 @@ sub new
     my $class = shift;
     my $self = {};
     my $description = shift;
+    my $save;
+    my $savedest;
 
     $self->{JobDescription} = $description;
 
-    #parse the XML blob that is the extensions element in the job description
-    if ($description->xml_extensions())
-    {
-        new Globus::GRAM::ExtensionsHandler($class, $description);
-    }
-
-    if(defined($description->logfile()))
-    {
-        local(*FH);
-        open(FH, '>>'. $description->logfile());
-        select((select(FH),$|=1)[$[]);
-        $self->{log} = *FH;
-    }
-
     bless $self, $class;
 
-    $self->log("New Perl JobManager created.");
     eval { File::Path::mkpath($self->job_dir(), 0, 0700); };
 
     if ($@) {
         $self->log("Couldn't create job dir");
     }
 
+    $save = $description->save_job_description();
+    $self->log("Checking to see if we'll save the job description: $save");
+
+    if ($save eq 'yes')
+    {
+        $savedest = "$ENV{HOME}/gram_".$description->uniq_id().".pl";
+        $self->log("Saving job description to $savedest");
+        $description->save($savedest);
+    }
+
+    $self->rewrite_urls();
+
     return $self;
 }
 
@@ -125,12 +123,13 @@ a timestamp.
 sub log
 {
     my $self = shift;
+    my $msg = join("", @_);
 
-    if(exists($self->{log}))
-    {
-        my $fh = $self->{log};
-	print $fh scalar(localtime(time)), " JM_SCRIPT: ", @_, "\n";
-    }
+    $msg =~ s/\\/\\\\/g;
+    $msg =~ s/\n/\\n/g;
+    $msg =~ s/\"/\\\"/g;
+
+    $self->respond({LOG => "msg=\"$msg\""});
 
     return;
 }
@@ -358,7 +357,7 @@ sub make_scratchdir
     {
         # Files with names comprised of Ascii values 48-122 should be
 	# relatively easy to remove from the shell if things go bad.
-	$tmpname = 'gram_scratch_' .
+	my $tmpname = 'gram_scratch_' .
 	           $acceptable[rand() * $#acceptable] .
 	           $acceptable[rand() * $#acceptable] .
 	           $acceptable[rand() * $#acceptable] .
@@ -481,6 +480,7 @@ sub rewrite_urls
     my $tag = $description->cache_tag() || $ENV{'GLOBUS_GRAM_JOB_CONTACT'};
     my $url;
     my $filename;
+    my $filestreamout = [];
 
     foreach ('stdin', 'executable')
     {
@@ -495,6 +495,73 @@ sub rewrite_urls
 	    }
 	}
     }
+    foreach my $which ('stdout', 'stderr')
+    {
+        my @destinations = $description->$which();
+        my $first_destination = $destinations[0];
+        my $first_tag = undef;
+        my $cached_destination = $self->job_dir() . "/$which";
+
+        if (ref($first_destination))
+        {
+            if (scalar(@{$first_destination}) == 2)
+            {
+                $first_tag = $first_destination->[1];
+                $first_destination = $first_destination->[0];
+                $self->log("destination = $first_destination, tag is $first_tag");
+            }
+            else
+            {
+                $first_destination = $first_destination->[0];
+                $self->log("destination = $first_destination, tag is not present");
+            }
+        }
+        elsif (scalar(@destinations) == 2)
+        {
+            $first_tag = $destinations[1];
+            $self->log("destination is $first_destination, tag is $first_tag");
+            @destinations = ($first_destination);
+        }
+
+        if (scalar(@destinations) == 1 && $first_destination !~ m|://|)
+        {
+            $description->add($which, $first_destination);
+            next;
+        }
+        elsif (scalar(@destinations) == 1 &&
+               $first_destination =~ m|x-gass-cache://|)
+        {
+	    my @arg = ($cache_pgm, '-add', '-t', $tag, '-n',
+                        $first_destination, 'file:///dev/null');
+            if (!defined($first_tag))
+            {
+                $self->log("$which goes to $first_destination in cache");
+                $self->log("command is " . join(" ", @arg));
+                $filename = $self->pipe_out_cmd(@arg);
+                if ($! != 0)
+                {
+                    $self->log("Error adding to cache $!");
+                }
+                $first_tag = $tag;
+            }
+
+	    @arg = ($cache_pgm, '-query', '-t', $first_tag, $first_destination);
+            $filename = $self->pipe_out_cmd(@arg);
+	    if($filename ne '')
+	    {
+                $self->log("$first_destination in cache resolves to $filename");
+		$description->add($which, $filename);
+	    }
+            else
+            {
+                $self->log("$first_destination in cache did not resolve!");
+            }
+        }
+        else
+        {
+            $description->add($which, $cached_destination);
+        }
+    }
     return {};
 }
 
@@ -528,7 +595,7 @@ sub stage_in
     {
         @arg = ($cache_pgm, '-add', '-t', $tag, $description->executable());
 
-        ($stderr, $rc) = $self->pipe_err_cmd(@arg);
+        ($stderr, $rc) = $self->pipe_out_cmd(@arg);
 
         if ($rc != 0) {
             $self->log("executable staging failed with $stderr");
@@ -680,16 +747,100 @@ sub stage_out
     my $description = $self->{JobDescription};
     my $url_copy = "$Globus::Core::Paths::bindir/globus-url-copy";
     my $tag = $description->cache_tag() || $ENV{'GLOBUS_GRAM_JOB_CONTACT'};
-    my $local_path;
+    my ($local, $remote);
+    my ($local_path, $remote_path);
+    my ($stderr, $rc);
     my @arg;
 
     $self->log("stage_out(enter)");
 
+    if (exists($self->{STDIO_MERGER}) && ref($self->{STDIO_MERGER}))
+    {
+        my $merger = $self->{STDIO_MERGER};
+        $merger->poll(1);
+    }
+
+
     $self->nfssync( $description->stdout(), 0 )
 	if defined $description->stdout();
     $self->nfssync( $description->stderr(), 0 )
 	if defined $description->stderr();
 
+    foreach ($description->file_stream_out())
+    {
+        next unless defined $_;
+
+	($local, $remote) = ($_->[0], $_->[1]);
+
+        $self->log("Staging $local to $remote");
+	# handle a couple of types of URLs for local files
+	$local_path = $local;
+	if($local_path =~ m|^x-gass-cache://|)
+	{
+            @arg = ($cache_pgm, '-query', '-t', $tag, $local_path);
+            $local_path = $self->pipe_out_cmd(@arg);
+
+            return Globus::GRAM::Error::STAGE_OUT_FAILED
+                if($local_path eq '');
+	}
+	elsif($local_path =~ m|^file:/|)
+	{
+	    $local_path =~ s|^file:/+|/|;
+	}
+	if($local_path !~ m|^/|)
+	{
+	    $local_path = $description->directory() . '/' . $local;
+	}
+
+        $self->nfssync($local_path, 0);
+
+        $remote_path = $remote;
+        if ($remote_path =~ m|^x-gass-cache://|)
+        {
+            my $msg;
+
+            @arg = ($cache_pgm, '-add',
+                    '-n', $remote_path,
+                    '-t', $tag,
+                    'file:///dev/null');
+            $msg = $self->pipe_out_cmd(@arg);
+            if ($? != 0)
+            {
+                $self->log("Failed creating cache entry for $remote_path $msg");
+                return Globus::GRAM::Error::STAGE_OUT_FAILED;
+            }
+            @arg = ($cache_pgm, '-query', '-t', $tag, $remote_path);
+            $remote_path = $self->pipe_out_cmd(@arg);
+            if ($? != 0)
+            {
+                $self->log("Query of cache URL path returned $? -- {$remote_path}");
+            }
+
+            return Globus::GRAM::Error::STAGE_OUT_FAILED
+                if ($remote_path eq '');
+            $remote_path = "file://$remote_path";
+        }
+
+        @arg = ($url_copy_pgm, 'file://' . $local_path, $remote_path);
+        $self->log("Staging $local to $remote");
+        $self->log("Concretely, staging $local_path to $remote_path");
+        $self->log("with command " . join(" ", @arg));
+
+        ($stderr, $rc) = $self->pipe_err_cmd(@arg);
+        if($rc != 0) {
+            $self->log("filestageout staging failed with $stderr");
+
+            $self->respond( {
+                'GT3_FAILURE_TYPE' => 'filestageout',
+                'GT3_FAILURE_MESSAGE' => $stderr,
+                'GT3_FAILURE_SOURCE' => $local,
+                'GT3_FAILURE_DESTINATION' => $remote
+            });
+            return Globus::GRAM::Error::STAGE_OUT_FAILED
+        }
+
+	$self->respond({'STAGED_STREAM' => "$local $remote"});
+    }
     foreach ($description->file_stage_out())
     {
         next unless defined $_;
@@ -749,6 +900,7 @@ sub cache_cleanup
     my $description = $self->{JobDescription};
     my $tag = $description->cache_tag() || $ENV{'GLOBUS_GRAM_JOB_CONTACT'};
     my $job_path = $self->job_dir();
+    my ($stderr, $rc);
 
     $self->log("cache_cleanup(enter)");
 
@@ -1015,7 +1167,7 @@ sub job_dir {
         if ($posix_hostname !~ m/\./) {
             my $aliases = join(' ',(gethostbyname($posix_hostname))[0,1]);
 
-            for $alias (split(/\s+/, $aliases)) {
+            for my $alias (split(/\s+/, $aliases)) {
                 if ($alias =~ m/\./) {
                     $posix_hostname = $alias;
 
@@ -1047,6 +1199,7 @@ sub setup_softenv
     my $soft_msc = shift;
     my $softenv_load = shift;
     my $job_script_fh = shift;
+    my $rc;
 
     my $description = $self->{JobDescription};
 
@@ -1063,14 +1216,22 @@ sub setup_softenv
         $self->log("default software environment requested");
 
         #load default software environment
-        print $job_script_fh ". $softenv_load\n";
+        $rc = print $job_script_fh ". $softenv_load\n";
+        if (!$rc)
+        {
+            return "print failed: $job_script_fh: $!";
+        }
     }
     else
     {
         $self->log("custom software environment requested");
 
         local(*SOFTENV);
-        open(SOFTENV, '>' . $softenv_script_name);
+        $rc = open(SOFTENV, '>' . $softenv_script_name);
+        if (!$rc)
+        {
+            return "open failed: $softenv_script_name: $!";
+        }
 
         foreach my $softenv (@softenv)
         {
@@ -1079,13 +1240,17 @@ sub setup_softenv
 
         close(SOFTENV);
 
-        print $job_script_fh "$soft_msc $softenv_script_name\n";
-        print $job_script_fh ". $softenv_script_name.cache.sh\n";
-        print $job_script_fh "rm $softenv_script_name"
-                           . " $softenv_script_name.cache.sh\n";
+        $rc = print $job_script_fh "$soft_msc $softenv_script_name\n"
+                                 . ". $softenv_script_name.cache.sh\n"
+                                 . "rm $softenv_script_name"
+                                 . " $softenv_script_name.cache.sh\n";
+        if (!$rc)
+        {
+            return "print failed: $job_script_fh: $!";
+        }
     }
 
-    return 1;
+    return 0;
 }
 
 1;
